/*------------------------------------------------------------------------------
#
#	Macintosh Developer Technical Support
#
#	MoreATSUnicodeTextBox.cp	-	Source file
#
#	Copyright  1998 Apple Computer, Inc.
#	All rights reserved.
#	
#	You may incorporate this sample code into your applications without
#	restriction, though the sample code has been provided "AS IS" and the
#	responsibility for its operation is 100% yours.  However, what you are
#	not permitted to do is to redistribute the source as "DSC Sample Code"
#	after having made changes. If you're going to re-distribute the source,
#	we require that you make it clear in the source that the code was
#	descended from Apple Sample Code, but that you've made changes.
#
#	Versions:	1.0					7/98
#
------------------------------------------------------------------------------*/



/*________ Includes ________________________________________________________________*/
#ifndef __FONTS__
#include <Fonts.h>	// GetSysFont(), GetAppFont()
#endif

#ifndef __SCRIPT__
#include <Script.h>	// GetSysDirection()
#endif

#ifndef __FIXMATH__
#include <FixMath.h>
#endif

#ifndef __TEXTEDIT__
#include <TextEdit.h>
#endif

#include "MoreATSUnicode.h"
#include "MoreATSUnicodeTextBox.h"
#include "MoreATSUnicodeInternal.h"

#define ff(a)			   	IntToFixed(a)
#define IntToFixed(a)	   	((Fixed)(a) << 16)


// 	converts a TextEdit based alignment to 
//	an ATSUI flush factor
static Fract atsuFlushFactorFromTEAlignment(short alignment)
{
	Fract flushFactor = kATSUStartAlignment;
	
	switch (alignment)
	{
		case teFlushRight:
			flushFactor = kATSUEndAlignment;
			break;
			
		case teCenter:
			flushFactor = kATSUCenterAlignment;
			break;

		case teFlushLeft:
			flushFactor = kATSUStartAlignment;
			break;
		default:
			// teFlushDefault
			flushFactor = (GetSysDirection() == 0) ? kATSUStartAlignment : kATSUEndAlignment;
			break;
	}
	return flushFactor;
}


enum {
		kVerticalTextAngle = ff(-90)
	};

/*
	Draw the text layout within the specified rect.
	
*/
OSStatus atsuDrawTextInBox(	ATSUTextLayout 		iTextLayout,
							const Rect *			iBox)
{
	OSStatus 			theStatus = noErr;
	UniCharArrayOffset	textOffset = 0;
	UniCharCount		textLength = 0;
	UniCharArrayOffset*	lineEndOffsets = NULL;
	
	require (iBox != NULL, EXIT);
	
	// the the range of text to be drawn
	theStatus = ATSUGetTextLocation (iTextLayout, NULL, NULL, &textOffset, &textLength, NULL);
	if (theStatus == noErr)
	{
		UniCharArrayOffset		lineStartOffset = textOffset;
		UniCharArrayOffset		lineEndOffset = 0;
		
		//	assume horizontal text values
		ATSUTextMeasurement		xPos = ff(iBox->left);
		ATSUTextMeasurement		yPos = ff(iBox->top);
		ATSUTextMeasurement		lineStart = xPos;
		ATSUTextMeasurement		lineEnd = ff(iBox->right);
		ATSUTextMeasurement		lineWidth = 0;
		
		Fixed					textAngle = 0;
		ItemCount				lineCount = 0;
		ItemCount				softBreakCount = 0;
		ATSUTextMeasurement		maxAscent = 0, maxDescent = 0;
		
		int ln = 0;

		//	need to get the text rotation for drawing
		theStatus = atsuGetLineRotation(iTextLayout, &textAngle);
		require (theStatus == noErr, EXIT);
		
		// is the text not horizontal?
		if (textAngle != 0)
		{
			// if it is not horizontal, we only know how to draw vertical
			require (textAngle == kVerticalTextAngle, EXIT);
			xPos = 	ff(iBox->right);
			lineStart = yPos;
			lineEnd = ff(iBox->bottom);
		}

		// check for linewidth set as a layout control
		theStatus = atsuGetLineWidth( iTextLayout, &lineWidth);
		require(theStatus == noErr, EXIT);
		
		//	if there is no layout control set for width
		//	then set it using the box bounds
		if (lineWidth == 0)
			lineWidth = lineEnd-lineStart;
		
		//	Break and measure each line to determine the max ascent and descent;
		//	This is needed because things that end up on different lines
		//	could interact to affect the line height when they're on the same line.
		//	To get the best value we break the lines first, then measure each
		//	line and find the max ascent and descent.
		while (lineStartOffset < textOffset+textLength)
		{
			ATSUTextMeasurement		ascent = 0, descent = 0;
			
			// set the soft breaks, we will use them later
			theStatus = ATSUBreakLine(iTextLayout, lineStartOffset, lineWidth, true, &lineEndOffset);
			require(theStatus == noErr, EXIT);
			
			theStatus = ATSUMeasureText(iTextLayout, lineStartOffset, lineEndOffset-lineStartOffset, NULL, NULL, &ascent, &descent);
			require(theStatus == noErr, EXIT);

			if (ascent > maxAscent)
				maxAscent = ascent;
			if (descent > maxDescent)
				maxDescent = descent;
				
			lineStartOffset = lineEndOffset;
			lineCount++;
			
		}
		
		lineEndOffsets = (UniCharArrayOffset*) NewPtr(lineCount * sizeof(UniCharArrayOffset));
		
		theStatus = MemError();
		require(theStatus == noErr, EXIT);
		
		theStatus = ATSUGetSoftLineBreaks(
			iTextLayout, textOffset, textLength, lineCount, lineEndOffsets, &softBreakCount);
		//	assert that the number of softbreaks is always one less than the number of lines
		//	since ATSUBreakLine does not insert a softbreak at the end of the text.
		require((theStatus == noErr && softBreakCount == lineCount-1), EXIT);

		lineEndOffsets[softBreakCount] = textOffset+textLength;
		lineStartOffset = textOffset;

		//	draw each line
		for (ln = 0; ln < lineCount; ln++)
		{
			lineEndOffset = lineEndOffsets[ln];
			
			// predecrement to account for the line height
			if (textAngle == kVerticalTextAngle)
			{
				xPos -= maxAscent;
				if (xPos < ff(iBox->left))
					break;
			} else {
				yPos += maxAscent;
				if (yPos > ff(iBox->bottom))
					break;
			}

			theStatus = ATSUDrawText(iTextLayout, lineStartOffset, lineEndOffset-lineStartOffset, xPos,  yPos );
			require(theStatus == noErr, EXIT);

			lineStartOffset = lineEndOffset;
			
			if (textAngle == kVerticalTextAngle)
			{
				xPos -= maxDescent;
			} else {
				yPos += maxDescent;
			}
		}
	}
	
EXIT:
	if (lineEndOffsets != NULL) 
		DisposePtr((char*)lineEndOffsets);	
	return theStatus;
}


/*
	atsuTextBox expands the functionality of TETextBox by adding
	justification and vertical text.
	
	atsuTextBox first sets up a ATSUTextLayout. It then draws the text
	using atsuDrawTextInBox
	
*/

enum {
	kInitialStyleAllocation = 10
	};

OSStatus atsuTextBox(	ConstUniCharArrayPtr 		iText,
						UniCharCount 				iTextLength,
						const Rect *				iBox,
						short 						alignment,
						Fract						justification,
						short						textOrientation,
						Boolean						useFontFallBacks)
{
	OSStatus		theStatus = 0;
	ATSUStyle		globalStyle = NULL;
	ATSUTextLayout	txLayout = NULL;
	
	int				substituteStyleCount = 0;
	int				maxStyleCount = kInitialStyleAllocation;
	ATSUStyle**		substituteStyles = NULL;
	
	theStatus = ATSUCreateStyle(&globalStyle);
	if (theStatus == noErr) 
	{
		//	Initialize the style with the current port attributes
		theStatus = atsuSetStyleFromGrafPtr(globalStyle, NULL, false);
		require(theStatus == noErr, EXIT);

		//	Don't want to have hanging punctuation drawn outside of the rectangle,
		//	so inhibit that feature. (It's allowed by default)
		theStatus = atsuSetHangingInhibitFactor(globalStyle, ((Fract) 0x00000001L));
		
		if (theStatus == noErr) 
		{
		
			theStatus = ATSUCreateTextLayoutWithTextPtr( iText, 0, iTextLength, iTextLength,
													1, &iTextLength, &globalStyle,
													&txLayout);
													
			if (theStatus == noErr) //	Set layout controls and draw
			{
				UniCharArrayOffset 	replaceOffset = 0;
				UniCharCount		replaceLength = iTextLength;
				
				//	Could have had atsuTextBox use the ATSUI flushFactors directly
				//	This is just to illustrate how to convert from the
				//	TE equivalents
				Fract	flushFactor = atsuFlushFactorFromTEAlignment(alignment);
				
				//	a slight optimization since we know that this
				//	text is display only
				theStatus = atsuSetLineLayoutOptions(txLayout, kATSLineIsDisplayOnly);
				require(theStatus == noErr, EXIT);

				//	Set the flush and justification if they are other than the 
				//	default
				if (flushFactor != kATSUStartAlignment || justification != kATSUNoJustification)
				{
					Fixed	lineWidth = ff(iBox->right-iBox->left);
					
					if (textOrientation == kVerticalText)
						lineWidth = ff(iBox->bottom-iBox->top);
									
					//	line width must be set for flush and justification to work
					theStatus = atsuSetLineWidth(txLayout, lineWidth);				
					require(theStatus == noErr, EXIT);

					theStatus = atsuSetLineFlushFactor(txLayout, flushFactor);		
					require(theStatus == noErr, EXIT);
							
					theStatus = atsuSetLineJustFactor(txLayout, justification);
					require(theStatus == noErr, EXIT);
					
					
					// don't justify the last line
					theStatus = atsuSetLineLayoutOptions(txLayout, kATSLineLastNoJustification);
					require(theStatus == noErr, EXIT);
				}
				
				if (textOrientation == kVerticalText)
				{
					// set the character orientation
					theStatus = atsuSetVerticalCharacter(globalStyle, kATSUStronglyVertical);
					require(theStatus == noErr, EXIT);
					
					//	rotate the text
					theStatus = atsuSetLineRotation(txLayout, kVerticalTextAngle);
					require(theStatus == noErr, EXIT);
				}

				//	If the caller specifies useFontFallBacks
				//	Check to see if there are any characters that can't be drawn
				//	with the font that came in from the port.
				//	Use ATSUMatchFontsToText to find problem subranges
				//	Explicitly replacing the font will be less than the overhead
				//	of setting transient font matching for the entire layout.
				if (useFontFallBacks)
				{
					Boolean	needTransientMatching = false;
					Boolean	doneWithSubstitution = false;
					
					while (!doneWithSubstitution)
					{
						ATSUFontID	replaceFont;
						
						theStatus = ATSUMatchFontsToText(
										txLayout, replaceOffset, replaceLength,
										&replaceFont, &replaceOffset, &replaceLength);
										
						switch (theStatus)
						{
							case kATSUFontsMatched:
							{
								int			i = 0;
								ATSUStyle	replacementStyle = NULL;
								
								//	"kATSUFontsMatched" means that some of the characters cannot
								//	be drawn with the font, but a suitable subsitute font
								//	is active in the system
								
								//	before we create a new style, see if we already have a style with this font
								for (i = 0; i < substituteStyleCount && !replacementStyle; i++)
								{
									ATSUFontID	thisFont;
									atsuGetFont((*substituteStyles)[i], &thisFont);
									if (thisFont == replaceFont)
									{
										replacementStyle = (*substituteStyles)[i];
									}
								}
								
								if (!replacementStyle)
								{
									//	need to create a new one, cloning the global style
									theStatus = ATSUCreateAndCopyStyle(globalStyle, &replacementStyle);
									require(theStatus == noErr, EXIT);
									
									theStatus = atsuSetFont(replacementStyle, replaceFont);
									require(theStatus == noErr, EXIT);
									
									//	add it to the list
									if (substituteStyles == NULL)
									{
										maxStyleCount = kInitialStyleAllocation;
										substituteStyles = (ATSUStyle**) NewHandle(maxStyleCount * sizeof(ATSUStyle));
										HLock((Handle) substituteStyles);
										require ((substituteStyles != NULL && MemError() == noErr), EXIT);
									} else {
										// do we need to expand our array of styles?
										if (substituteStyleCount > maxStyleCount)
										{
											maxStyleCount += kInitialStyleAllocation;
											HUnlock((Handle) substituteStyles);
											SetHandleSize((Handle) substituteStyles, maxStyleCount * sizeof(ATSUStyle));
											HLock((Handle) substituteStyles);
											theStatus = MemError();
											require(theStatus == noErr, EXIT);
										}
									}
									(*substituteStyles)[substituteStyleCount++] = replacementStyle;
								}

								theStatus = ATSUSetRunStyle(txLayout, replacementStyle, replaceOffset, replaceLength);
								require (theStatus == noErr, EXIT);
								break;
							}
							case kATSUFontsNotMatched:
							{
								//	kATSUFontsNotMatched" means that some of the characters cannot
								//	be drawn with any currently active font.
								//	this will tell us to turn on transient font matching 
								//	and pick up glyphs from the last resort font 
								needTransientMatching = true;
								break;
							}
							default:
								//	done with any other status
								doneWithSubstitution = true;
								break;
						}
						replaceOffset += replaceLength;
						replaceLength = iTextLength - replaceOffset;
					} 
					require(theStatus == noErr, EXIT);
					
					if (needTransientMatching)
					{
						theStatus = ATSUSetTransientFontMatching(txLayout, true);
						require(theStatus == noErr, EXIT);
					}
				}
				
				//	Finally, this is where the drawing happens
				theStatus = atsuDrawTextInBox(txLayout, iBox);
			}
		}
	}

EXIT:
	if (globalStyle)
	{
		ATSUDisposeStyle(globalStyle);
	}
	
	if (substituteStyles)
	{
		int i;
		for (i = 0; i < substituteStyleCount; i++)
		{
			ATSUDisposeStyle((*substituteStyles)[i]);
		}
		HUnlock((Handle) substituteStyles);
		DisposeHandle((Handle) substituteStyles);	

	}
	if (txLayout) {
		ATSUDisposeTextLayout(txLayout);
	}
	return theStatus;
}


